@@ -16,12 +16,7 @@ from __future__ import absolute_import
 __author__ = 'afshar@google.com (Ali Afshar)'
 
 
-# Oauth2client < 3 has the positional helper in 'util', >= 3 has it
-# in '_helpers'.
-try:
-  from oauth2client import util
-except ImportError:
-  from oauth2client import _helpers as util
+from googleapiclient import _helpers as util
 
 
 def setup_package():
